/*  detect.c
 *
 *  Signal detection functions of xdemorse application
 */

/*
 *  xdemorse: An application to decode Morse code signals to text
 *
 *  Copyright (C) 2002  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


#include "detect.h"

/* Runtime config data */
extern rc_data_t rc_data;

extern gint gbl_scope_height;

/*------------------------------------------------------------------------*/

/*  Get_Fragment()
 *
 *  Detects the cw beat frequency signal o/p from the radio
 *  receiver and determines the status, (mark or space) of a
 *  'fragment' (a small fraction, ~1/8) of a Morse code element.
 *  Signal detection is done by using a Goertzel algorithm.
 */

  gboolean
Get_Fragment( void )
{
  int
	slope_ave,	/* Average value of slope of Geortzel o/p waveform */
	slope_pos,	/* Detector output slope is positive */
	slope_neg,	/* Detector output slope is negative */
	slope_zero,	/* Detector output slope is zero */
	block_size, /* Block size (N) of the Goertzel algorithm */
	frag_timer, /* Counter for timing duration of fragment  */
	frag_level,	/* Level of audio signal over a fragment */
	prev_frag_level, /* Previous value of above */
	idx, ref;

  /* Variables for the Goertzel algorithm */
  double q0, q1, q2, level;

  /* Background (average) value of average slope */
  static int slope_bkg = 0;

  char slope_status;

  /* Goertzel block size depends on Morse speed */
  block_size = detector_data.frag_len * rc_data.unit_elem;

  /* Buffer dsp samples of input signal for a fragment */
  for( frag_timer = 0; frag_timer < detector_data.frag_len; frag_timer++ )
  {
	if( !Signal_Sample(
		  &detector_data.samples_buff[detector_data.samples_buff_idx]) )
	return( FALSE );

	/* Advance/Reset circular buffers' index */
	detector_data.samples_buff_idx++;
	if( detector_data.samples_buff_idx >= detector_data.samples_buff_len )
	  detector_data.samples_buff_idx = 0;
  } /* for( frag_timer = 0; frag_timer < frag_len ... */

  /** Calculate signal fragment level over a block **/
  /* Backstep buffer index for use of samples */
  detector_data.samples_buff_idx -= block_size;
  if( detector_data.samples_buff_idx < 0 )
	detector_data.samples_buff_idx += detector_data.samples_buff_len;

  /* Calculate fragment level using Goertzel algorithm */
  q1 = q2 = 0.0;
  for( idx = 0; idx < block_size; idx++ )
  {
	q0 = detector_data.coeff * q1 - q2 +
	  (double)detector_data.samples_buff[detector_data.samples_buff_idx];
	q2 = q1;
	q1 = q0;

	/* Advance/Reset circular buffers' index */
	detector_data.samples_buff_idx++;
	if( detector_data.samples_buff_idx >= detector_data.samples_buff_len )
	  detector_data.samples_buff_idx = 0;
  } /* for( idx = 0; idx < block_size; idx++ ) */

  /* Scalar magnitude of input signal scaled by block size */
  q1 /= (double)block_size;
  q2 /= (double)block_size;
  level = q1*q1 + q2*q2 - q1*q2*detector_data.coeff;
  frag_level = (int)sqrt( level );

  /* Save signal power level to circular buffer */
  detector_data.sig_level_buff[detector_data.sig_level_idx] = frag_level;

  /* Advance/Reset circular buffers' index */
  detector_data.sig_level_idx++;
  if( detector_data.sig_level_idx >= rc_data.max_unit )
	detector_data.sig_level_idx = 0;

  /* Backstep buffer index for use of fragment levels */
  detector_data.sig_level_idx -= rc_data.unit_elem;
  if( detector_data.sig_level_idx < 0 )
	detector_data.sig_level_idx += rc_data.max_unit;

  /* Calculate average value of slope of Goertzel o/p waveform */
  prev_frag_level = detector_data.sig_level_buff[detector_data.sig_level_idx];

  /* Advance/Reset circular buffers' index */
  detector_data.sig_level_idx++;
  if( detector_data.sig_level_idx >= rc_data.max_unit )
	detector_data.sig_level_idx = 0;

  /* Summate slope values */
  slope_ave = 0;
  for( idx = 1; idx < rc_data.unit_elem; idx++ )
  {
	int diff;
	frag_level = detector_data.sig_level_buff[detector_data.sig_level_idx];
	diff = frag_level - prev_frag_level;
	slope_ave += diff;
	prev_frag_level = frag_level;

	/* Advance/Reset circular buffers' index */
	detector_data.sig_level_idx++;
	if( detector_data.sig_level_idx >= rc_data.max_unit )
	  detector_data.sig_level_idx = 0;
  } /* for( idx = 1; idx < rc_data.unit_elem; idx++ ) */
  
  /* Asymptotic summation of slope average */
  slope_bkg =
	( abs(slope_ave) + (SLOPE_AVERAGE - 1) * slope_bkg ) / SLOPE_AVERAGE;

  /* Use slope background as detector threshold */
  rc_data.det_thresh = slope_bkg / THRESH_SCALE;
  if( rc_data.det_thresh < rc_data.min_det_thresh )
	rc_data.det_thresh = rc_data.min_det_thresh;

  /* Detect slope status */
  ref = rc_data.det_thresh * DETECT_SCALE;
  if( slope_ave > ref )
	slope_status = SLOPE_POSITIVE;
  else if( slope_ave < -ref )
	slope_status = SLOPE_NEGATIVE;
  else
	slope_status = SLOPE_ZERO;

  /* Save current slope status in circular buffer */
  detector_data.slope_status_buff[detector_data.slope_status_idx] =
	slope_status;
  detector_data.slope_status_idx++;
  if( detector_data.slope_status_idx >= rc_data.max_unit )
	detector_data.slope_status_idx = 0;

  /* Backstep buffer index for use of slope status */
  detector_data.slope_status_idx -= rc_data.unit_elem;
  if( detector_data.slope_status_idx < 0 )
	detector_data.slope_status_idx += rc_data.max_unit;

  /* Count the number of slope status states */
  slope_pos = slope_neg = slope_zero = 0;
  for( idx = 0; idx < rc_data.unit_elem; idx++ )
  {
	slope_status =
	  detector_data.slope_status_buff[detector_data.slope_status_idx];
	if( slope_status == SLOPE_POSITIVE )
	  slope_pos++;
	else if( slope_status == SLOPE_NEGATIVE )
	  slope_neg++;
	else
	  slope_zero++;

	detector_data.slope_status_idx++;
	if( detector_data.slope_status_idx >= rc_data.max_unit )
	  detector_data.slope_status_idx = 0;
  }

  /* Set signal tone detector status */
  if( (slope_pos > slope_zero) && (slope_pos > slope_neg) )
  {
	Set_Flag( MARK_TONE );
	Clear_Flag( SPACE_TONE );
  }
  else if( (slope_neg > slope_zero) && (slope_neg > slope_pos) )
  {
	Clear_Flag( MARK_TONE );
	Set_Flag( SPACE_TONE );
  }

  /* Display signal graph (scope) */
  if( isFlagSet(DISPLAY_DETECTOR) )
	Display_Detector( slope_ave / DETECT_SCALE );
  else if( isFlagSet(DISPLAY_SIGNAL) )
	  Display_Signal( frag_level / SIGNAL_SCALE );

  return( TRUE );
} /* End of Get_Fragment() */

/*------------------------------------------------------------------------*/

/* Initialize_Detector()
 *
 * Initializes variables and allocates buffers
 * for the Goertzel tone detector
 */
  gboolean
Initialize_Detector( void )
{
  size_t alloc;

  /* Goertzel detector coefficients */
  double w = 2.0 * M_PI *
	(double)rc_data.tone_freq /
	(double)rc_data.dsp_rate;
  detector_data.cosw  = cos(w);
  detector_data.sinw  = sin(w);
  detector_data.coeff = 2.0 * detector_data.cosw;

  /* Clear buffer indices */
  detector_data.samples_buff_idx = 0;
  detector_data.sig_level_idx    = 0;
  detector_data.slope_status_idx = 0;

  /* Lenght of a signal 'fragment', a small part of one Morse element */
  detector_data.frag_len =
	(rc_data.dsp_rate * CYCLES_PER_FRAG) / rc_data.tone_freq;

  /* Length of DSP samples buffer for Goertzel detector */
  detector_data.samples_buff_len =
	(CYCLES_PER_FRAG * rc_data.max_unit * rc_data.dsp_rate) / rc_data.tone_freq;

  /* Allocate samples buffer */
  alloc = detector_data.samples_buff_len * sizeof(short);
  if( !mem_alloc((void *)&detector_data.samples_buff, alloc) )
	return( FALSE );

  /* Allocate signal level (Goerstzel detector o/p) buffer */
  alloc = sizeof(int) * rc_data.max_unit;
  if( !mem_alloc( (void *)&detector_data.sig_level_buff, alloc) )
	return( FALSE );

  /* Allocate slope status buffer */
  if( !mem_alloc( (void *)&detector_data.slope_status_buff, alloc) )
	return( FALSE );

  return( TRUE );
} /* Initialize_Detector() */

/*------------------------------------------------------------------------*/

